@@ -26,9 +26,8 @@
 #include "goagoogleprovider.h"
 #include "goaimapsmtpprovider.h"
 #include "goaowncloudprovider.h"
+#include "goawebdavprovider.h"
 #include "goawindowsliveprovider.h"
-#include "goamediaserverprovider.h"
-#include "goalastfmprovider.h"
 
 #ifdef GOA_FEDORA_ENABLED
 #include "goafedoraprovider.h"
@@ -944,20 +943,17 @@ static struct
 #ifdef GOA_EXCHANGE_ENABLED
   { GOA_EXCHANGE_NAME, goa_exchange_provider_get_type },
 #endif
-#ifdef GOA_LASTFM_ENABLED
-  { GOA_LASTFM_NAME, goa_lastfm_provider_get_type },
-#endif
 #ifdef GOA_FEDORA_ENABLED
   { GOA_FEDORA_NAME, goa_fedora_provider_get_type },
 #endif
 #ifdef GOA_IMAP_SMTP_ENABLED
   { GOA_IMAP_SMTP_NAME, goa_imap_smtp_provider_get_type },
 #endif
+#ifdef GOA_WEBDAV_ENABLED
+  { GOA_WEBDAV_NAME, goa_webdav_provider_get_type },
+#endif
 #ifdef GOA_KERBEROS_ENABLED
   { GOA_KERBEROS_NAME, goa_kerberos_provider_get_type },
-#endif
-#ifdef GOA_MEDIA_SERVER_ENABLED
-  { GOA_MEDIA_SERVER_NAME, goa_media_server_provider_get_type },
 #endif
   { NULL, NULL }
 };
